from src.models.vessel import VesselRequest
from src.models.enums import AscDescEnum, AggregateType, ResolutionType, EncodingType
from src.models.cache_interface import CacheInterface
from src.models.database_interface import DatabaseInterface

from pandas import DataFrame, Timestamp
from datetime import datetime
from typing import Optional, Union, List, Any
import json


class VesselService:

    def __init__(self, db_context: DatabaseInterface, cache_context: CacheInterface):
        self.context = db_context
        self.cache = cache_context

    def _sanitize_field(self, fields: Union[str, List], available_fields: List[str]) -> bool:
        """

        Args:
            fields: List of fields to be sanitized
            available_fields: Fields available on the database

        Returns: Boolean value if fields are fetchable

        """
        if type(fields) == str:
            if not any(d['fieldKey'] == fields for d in available_fields):
                raise ValueError(f'Error parsing {fields}')

        for field in fields:
            if not any(d['fieldKey'] == field for d in available_fields):
                raise ValueError(f'Error parsing {field}')

        return True

    @staticmethod
    def _construct_select_query(measurement: str,
                                fields: str,
                                limit: Optional[int] = None,
                                offset: Optional[int] = None,
                                asc_desc: Optional[AscDescEnum] = None,
                                order_by: Optional[str] = 'time',
                                from_date: Optional[datetime] = None,
                                to_date: Optional[datetime] = None,
                                resolution_type: ResolutionType = None,
                                resolution: int = None
                                ) -> str:

        """
        Construct a SELECT statement to be sent to the database
        Args:
            measurement: Which measurement to select FROM
            fields: String of comma separated fields to be selected
            limit: Optional amount of rows to fetch
            offset: Optional row offset
            from_date: Optional Datetime object to find values from
            to_date: Optional Datetime object fo find values to

        Returns: String query ready to be sent to influx
        """
        query = f'SELECT {fields} FROM {measurement} '

        # From given date until now
        if from_date is not None and to_date is None and from_date is type(from_date) == datetime:
            query += f"WHERE time > '{from_date}' AND time < now()"

        # From given date till given date
        if from_date is not None and to_date is not None and type(from_date) == datetime and type(to_date) == datetime:
            query += f"WHERE time > '{from_date}' AND time < '{to_date}' "

        if resolution is not None and resolution_type is not None:
            query += f'GROUP BY time({resolution}{resolution_type.value}) FILL(none) '

        if asc_desc is not None and type(asc_desc) == AscDescEnum:
            query += f'ORDER BY {order_by} {asc_desc.value} '

        if limit is not None and type(limit) == int:
            query += f'LIMIT {limit} '

        if offset is not None and type(offset) == int:
            query += f'OFFSET {offset} '

        query += ';'
        return query

    def _parse_request_body_aggregates(self,
                                       request_body: Union[str, VesselRequest],
                                       aggregate_type: AggregateType
                                       ) -> str:
        """
        Parses request body for aggregated results
        Args:
            request_body: Request body from client
            aggregate_type: Which type of aggregation to apply to query

        Returns: A sanitized concatenated string of comma separated fields from request body with aggregation type applied.
        """
        def add_aggregate(field: str) -> str:
            return f'{aggregate_type.value}({field})'

        if type(request_body) is str:
            return f'{aggregate_type.value}({request_body})'

        # Sanitizes input by checking if fields actually exist
        columns = list(VesselService.get_columns(self))
        if self._sanitize_field(request_body.fields, columns):
            return f', '.join(list(map(add_aggregate, request_body.fields)))

        raise ValueError('Something went wrong when parsing fields...')

    def _parse_request_body(self, request_body: Union[str, VesselRequest]) -> str:
        """
        Parses request body for normal results
        Args:
            request_body: Request body from client

        Returns: A sanitized concatenated string of comma separated fields from request body

        """
        if type(request_body) is str:
            return request_body

        # Sanitizes input by checking if fields actually exist
        columns = list(self.get_columns())
        if self._sanitize_field(request_body.fields, columns):
            return ', '.join(request_body.fields)

        raise ValueError('Something went wrong when parsing fields...')

    def _reformat_and_encode_response(self,
                                      data: DataFrame,
                                      measurement: str,
                                      encoding_type: EncodingType
                                      ) -> List[List[Any]]:
        """
        Formats key-value dataframe to a list of values, with optional encoding
        Args:
            data: DataFrame containing query results from database
            measurement: Name of which measurement was queried
            encoding_type: Encoding type to be applied to time stamps

        Returns: The same data formatted in a more optimal way with optional encoding
        """
        field_names = ['timestamp']
        field_names += list(data[measurement].columns)

        response = [field_names]
        timestamps = list(data[measurement].axes[0].to_series())
        values = list(data[measurement].values)

        if encoding_type == EncodingType.NONE:
            response += self._no_encoding(timestamps, values)

        if encoding_type == EncodingType.DELTA:
            response += self._delta_encoding(timestamps, values)

        if encoding_type == EncodingType.DELTA_OF_DELTA:
            response += self._delta_of_delta_encoding(timestamps, values)
        return response

    def _delta_of_delta_encoding(self, timestamps: List[Timestamp], values: List[Any]):
        """
        Applies delta-delta encoding and formats data to desired list format
        Args:
            timestamps: List of pandas time stamp objects
            values: List of values to be associated with given time stamps

        Returns: List of lists with time stamp and values together

        """
        response = []
        prev_ts = int(timestamps[0].timestamp())
        prev_delta = -prev_ts

        for ts, val in zip(timestamps, values):
            time = int(ts.timestamp())
            delta = time - prev_ts
            delta_of_delta = delta - prev_delta

            response.append([delta_of_delta, *val])
            prev_ts = time
            prev_delta = delta

        return response

    def _delta_encoding(self, timestamps: List[Timestamp], values: List[Any]):
        """
        Applies delta encoding and formats data to desired list format
        Args:
            timestamps: List of pandas time stamp objects
            values: List of values to be associated with given time stamps

        Returns: List of lists with time stamp and values together

        """
        response = []
        prev_ts = 0

        for ts, val in zip(timestamps, values):
            time = int(ts.timestamp())
            delta = time - prev_ts

            response.append([delta, *val])
            prev_ts = time

        return response

    def _no_encoding(self, timestamps: List[Timestamp], values: List[Any]):
        """
        Formats data to desired list format without any encoding
        Args:
            timestamps: List of pandas time stamp objects
            values: List of values to be associated with given time stamps

        Returns: List of lists with time stamp and values together
        """
        response = []
        for ts, val in zip(timestamps, values):
            time = int(ts.timestamp())
            response.append([time, *val])
        return response

    def hash_request_arguments(self, args):
        """
        Args:
            args: Arguments from incoming request

        Returns: A unique hash based on parameters
        """
        args['request_body'] = tuple(args['request_body'].fields)
        return hash(frozenset(args.items()))

    def seed_all(self, measurement: str = 'testship'):
        """
        Imports all of the NetCDF data in the root netcdf folder
        Args:
            measurement: Name of measurement associated to the data
        """
        self.context.upload_all('/code/netcdf/', measurement)

    def seed(self, filename: str, measurement: str = 'testship'):
        """
        Imports a single NetCDF file
        Args:
            filename: Name of the to be imported file
            measurement: name of the measurement associated to the data
        """
        self.context.upload_values(f'/code/netcdf/{filename}', measurement)

    def get_values(
            self,
            measurement: str,
            request_body: Union[str, VesselRequest] = '*',
            from_date: Optional[datetime] = None,
            to_date: Optional[datetime] = None,
            asc_desc: AscDescEnum = None,
            order_by: Optional[str] = 'time',
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            encoding_type: EncodingType = None
            ) -> List[List[Any]]:
        """
        Args:
            measurement: Name of measurement to be queried
            request_body: Request body containing which fields to be queried
            from_date: ISO date to query to
            to_date: ISO date to query from
            asc_desc: Ascending or descending order
            order_by: Field to order by
            limit: Max rows to be fetched for each field
            offset: Offset to be applied to rows fetched
            encoding_type: Which encoding to apply to time stamps

        Returns: Formatted data from database
        """

        cache_hash = str(self.hash_request_arguments(locals()))
        data = self.cache.get(cache_hash)  # Set to None if you want to circumvent cache
        if data is not None:
            return json.loads(data)  # Fetches cached data as binary, parse to JSON before returning

        fields = self._parse_request_body(request_body)
        query = self._construct_select_query(measurement=measurement,
                                             fields=fields,
                                             limit=limit,
                                             offset=offset,
                                             asc_desc=asc_desc,
                                             order_by=order_by,
                                             to_date=to_date,
                                             from_date=from_date)
        data = self.context.query(query)
        response = self._reformat_and_encode_response(data, measurement, encoding_type)
        self.cache.set(cache_hash, json.dumps(response))
        return response

    def get_multiple_aggregates(self,
                                request_body: Union[str, VesselRequest],
                                measurement: str,
                                aggregate_type: AggregateType,
                                resolution_type: ResolutionType,
                                resolution: int,
                                to_date: Optional[datetime],
                                from_date: Optional[datetime],
                                limit: int,
                                offset: int,
                                encoding_type: EncodingType
                                ) -> List[List[Any]]:
        """
        Args:
            measurement: Name of measurement to be queried
            request_body: Request body containing which fields to be queried
            aggregate_type: Type of aggregation to be applied
            resolution_type: Unit of time
            resolution: Time unit to fetch data from
            to_date: ISO date to fetch data from
            from_date: ISO date to fetch data from
            limit: Max rows to be fetched per field
            offset: Static offset
            encoding_type: Time stamp encoding

        Returns: Formatted data from database
        """

        cache_hash = str(self.hash_request_arguments(locals()))
        data = self.cache.get(cache_hash)
        if data is not None:
            return json.loads(data)

        fields = self._parse_request_body_aggregates(request_body=request_body,
                                                     aggregate_type=aggregate_type)
        query = self._construct_select_query(measurement=measurement,
                                             fields=fields,
                                             limit=limit,
                                             offset=offset,
                                             from_date=from_date,
                                             to_date=to_date,
                                             resolution_type=resolution_type,
                                             resolution=resolution)

        data = self.context.query(query)

        response = self._reformat_and_encode_response(data, measurement, encoding_type)

        self.cache.set(cache_hash, json.dumps(response))
        return response

    def get_columns(self):
        """
        Returns: List of fields available on the database
        """
        data = self.context.query('SHOW field keys').get_points()
        return data
